local _, c = UnitClass("player")
if c ~= "ROGUE" then return end

local options
local lastSwapTime = 0
local modeOverride
local playerGUID
local pendingEquip
local lastMHAttack, lastMHAttackSpeed = 0, 0
local use_debug = false

local function l_debug(...)
	if use_debug then
		print(...)
	end
end

PoisonSwapper = LibStub("AceAddon-3.0"):NewAddon("PoisonSwapper", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0")
local mod = PoisonSwapper
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")

-- With Rank
local deadlyName, _, deadlyIcon = GetSpellInfo(57970)

-- Without rank
local instantName, _, instant_icon 		= GetSpellInfo(8680)
local woundName, _, wound_icon	 		= GetSpellInfo(13219)
local cripplingName, _, crippling_icon 	= GetSpellInfo(25809)
local mindNumbingName, _, mnIcon	 	= GetSpellInfo(25810)
local mnCheck = mindNumbingName:gsub("-", " ")
local icons = {
	[instantName] = instant_icon,
	[woundName] = wound_icon,
	[cripplingName] = crippling_icon,
	[mnCheck] = mindNumbingName,
	[deadlyName] = deadlyIcon
}

local tt = CreateFrame("GameTooltip", "PoisonSwapTooltip", UIParent, "GameTooltipTemplate")
tt:SetOwner(UIParent, "ANCHOR_NONE")

local ldbObj = LDB:NewDataObject("PoisonSwapper", {
	type = "data source",
	icon = deadlyIcon,
	label = "PoisonSwapper",
	text  = "|cff00ff00Enabled|r",	
	OnClick = function(self, btn)
		mod:Toggle()
	end
})

local function scanTooltipFor(poison)
	poison = poison:lower()
	for i = 1, tt:NumLines() do
		local l = _G["PoisonSwapTooltipTextLeft" .. i]
		if l then
			local text = l:GetText()
			if text then
				local poisonMatch = text:lower():match(poison)
				if poisonMatch then
					return true
				end
			end
		end
	end
	return false
end

local function scanOffhandFor(poison)
	if not GetInventoryItemLink("player", 17) then return end
	tt:SetInventoryItem("player", 17)
	return scanTooltipFor(poison)
end

local function scanFor(bag, slot, poison)
	tt:SetBagItem(bag, slot)
	return scanTooltipFor(poison)
end

local function checkForPoison(targetName, targetStacks, timeLeft)
	local stacks, expires, name, unit
	for i = 1, 40 do 
		name, _, _, stacks, _, _, expires, unit = UnitDebuff("target", i)
		if name == targetName and unit == "player" then 					
			break
		end
		if not name then break end
	end
	if expires then
		return expires - GetTime() > timeLeft and (targetStacks == 1 or stacks >= targetStacks)
	end
end

function mod:StartCooldown()
	lastSwapTime = GetTime()
end

function mod:CooledDown()
	return true
	-- return GetTime() - lastSwapTime > options.Cooldown
end

local swappedAt = 0
function mod:Equip(ofType)
	mod:ExecuteSwap()
	if pendingEquip == ofType then
		l_debug("Pending equip request:", pendingEquip)
		return
	end
	l_debug("Swap requested to", ofType)
	
	-- Don't swap if we already have the desired poison type equipped.
	if scanOffhandFor(ofType) then
		l_debug("Offhand already has a weapon with", ofType, ", aborting swap")
		return
	end
	
	swappedAt = GetTime()
	pendingEquip = ofType
	ldbObj.text = ofType
	ldbObj.icon = icons[ofType]
	if GetTime() - lastMHAttack < 0.1 then
		self:ExecuteSwap(true)
	end
end

function mod:ExecuteSwap(isMainhand)
	-- If we've had a request pending for >= 3 seconds, force the swap. We missed a mainhand swing somewhere, likely due to a haste change.
	-- Don't swap if we have enough energy for a Mutilate (or less costly move). We don't want to interfere with the GCD and cause energy capping.

	local energyGood = true -- UnitMana("player") < 60	
	if pendingEquip and energyGood and (isMainhand or GetTime() - swappedAt > 3) then
		-- Sanity check, in case offhand changes between our equip request and it actually being executed.
		if scanOffhandFor(ofType) then
			l_debug("Offhand already has a weapon with", ofType, ", aborting swap")
			pendingEquip, swappedAt = nil, nil
			return
		end	
		
		l_debug("Swapping, weapon swing timer reset")
		local location = self:FindWeapon(pendingEquip)
		if location then
			EquipmentManager_EquipContainerItem(EquipmentManager_EquipItemByLocation(location, 17))
		else
			l_debug("Couldn't find any weapon with", ofType, ", aborting swap")
		end
		pendingEquip, swappedAt = nil, nil
	end
end

function mod:HasImpPoisons()
	-- if self.hasImpPoisons ~= nil then return self.hasImpPoisons end
	local _, _, _, _, points = GetTalentInfo(1, 11)
	return points >= 3
	-- return self.hasImpPoisons
end

local function GetRealTimeLeft()
	local realTimeLeft = lastMHAttackSpeed - (GetTime() - lastMHAttack)
	if realTimeLeft < 0 then realTimeLeft = 0 end
	return realTimeLeft + options.TimeLeft
end

function mod:PVESwap()
	local realTimeLeft = GetRealTimeLeft()
	local useDamage = checkForPoison(deadlyName, 5, realTimeLeft)
	if useDamage then
		if self:HasImpPoisons() then			
			self:Equip(instantName)
		else
			self:Equip(woundName)
		end
	else
		self:Equip(deadlyName)
	end
end

function mod:PVPSwap()
	local _, targetClass = UnitClass("target")
	local powerType = UnitPowerType("target")
	local useMN, useDamage
	local _, _, _, _, points = GetTalentInfo(1, 18)
	local hasDeadlyBrew = points > 0
	
	local realTimeLeft = GetRealTimeLeft()
	
	if powerType == 0 and targetClass ~= "HUNTER" then
		useMN = hasDeadlyBrew or checkForPoison(cripplingName, 1, realTimeLeft)
		if useMN then
			useDamage = checkForPoison(mindNumbingName, 1, realTimeLeft)
		end
	else
		useDamage = hasDeadlyBrew or checkForPoison(cripplingName, 1, realTimeLeft)
	end	
	if useDamage and self:CooledDown() then
		if self:HasImpPoisons() then
			self:Equip(instantName)
		else
			self:Equip(woundName)
		end
	elseif useMN and self:CooledDown() then
		self:Equip(mnCheck)
	else
		self:StartCooldown()
		self:Equip(cripplingName)
	end
end

function mod:CheckSwap()
	local pvp = (modeOverride ~= "pve") and (modeOverride == "pvp" or UnitIsPVP("target") and UnitCanAttack("player", "target"))
	if pvp then
		self:PVPSwap()
	else
		self:PVESwap()
	end
end

local buffer = {}
function mod:FindWeapon(ofType)
	wipe(buffer)
	GetInventoryItemsForSlot(17, buffer)
	for location, itemID in pairs(buffer) do
		l_debug("Checking:", location, itemID)
		local player, bank, bags, slot, bag = EquipmentManager_UnpackLocation(location)
		if bags and bag then
			if scanFor(bag, slot, ofType) then
				return location
			end
		end
	end
	l_debug("Done checking inventory")
	return nil
end

function mod:HandleEvent(event)
	if not PoisonSwapperEnabled then return end
	
	if event ~= "PLAYER_TARGET_CHANGED" and not InCombatLockdown() then return end
	if not UnitCanAttack("player", "target") then return end
	
	self:CheckSwap()
	
	if event == "PLAYER_TARGET_CHANGED" then
		mod:ExecuteSwap(true)
	end
end

function mod:OnInitialize()
	if PoisonSwapperEnabled == nil then PoisonSwapperEnabled = true end
	PoisonSwapperOptions = PoisonSwapperOptions or {}
	options = PoisonSwapperOptions
	options.TimeLeft = options.TimeLeft or 5
	options.Cooldown = options.Cooldown or 1
	
	-- self:SetEnabledState(PoisonSwapperEnabled)
	self:Toggle(PoisonSwapperEnabled)
	SLASH_POISONSWAPPER1 = "/ps"
	SlashCmdList["POISONSWAPPER"] = mod.SlashCommand
end

function mod.SlashCommand(msg)
	local cmd, other = strsplit(" ", msg, 2)
	other = other or ""
	cmd = cmd:lower()
	if cmd == "pvp" or cmd == "pve" or cmd == "auto" then
		print(("PoisonSwapper is now in |cffffff00%s mode|r"):format(cmd))
		modeOverride = cmd		
		return	
	elseif cmd == "time" and tonumber(other) > 0 then
		options.TimeLeft = tonumber(other)
		print(("PoisonSwapper will now swap to high-priority poisons at |cffffff00%s seconds left|r on debuff."):format(options.TimeLeft))
		return
	elseif cmd == "cooldown" and tonumber(other) >= 0 then
		options.Cooldown = tonumber(other)
	elseif cmd == "debug" then
		use_debug = not use_debug
		print(("PoisonSwapper debug is now |cffffff00%s|r."):format(use_debug and "enabled" or "disabled"))
		return
	end
	
	mod:Toggle()
end

function mod:Toggle(state)
	PoisonSwapperEnabled = state == nil and (not PoisonSwapperEnabled) or state
	local color, state
	if PoisonSwapperEnabled then
		color, state = "ff00ff00", "Enabled"
	else
		color, state = "ffff0000", "Disabled"
	end
	ldbObj.text = ("|c%s%s|r"):format(color, state)
	print(("PoisonSwapper: |c%s%s|r"):format(color, state))
end

function mod:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "HandleEvent")
	self:RegisterBucketEvent("UNIT_COMBAT", 0.1, "HandleEvent")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	playerGUID = UnitGUID("player")
end

local attackTimes, attackSpeeds = {}, {}
local tinsert, tremove, wipe = _G.tinsert, _G.tremove, _G.wipe
local UnitAttackSpeed = _G.UnitAttackSpeed
function mod:COMBAT_LOG_EVENT_UNFILTERED(evt, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	if sourceGUID ~= playerGUID then return end
	if event == "SWING_DAMAGE" or event == "SWING_MISSED" then
		local mainSpeed, offSpeed = UnitAttackSpeed("player")
		lastMHAttackSpeed = mainSpeed
		
		local isMainHand = false
		for i = 1, #attackTimes do
			local t = attackTimes[i]
			if not t then break end
			local delta = timestamp - t
			local expectedDelta = attackSpeeds[i]
			if delta > expectedDelta - 0.15 and delta < expectedDelta + 0.15 then
				isMainHand = true
				lastMHAttack = GetTime()
				wipe(attackTimes)
				wipe(attackSpeeds)
				break
			end
		end
		
		tinsert(attackTimes, 1, timestamp)
		tinsert(attackSpeeds, 1, mainSpeed)
		if #attackTimes > 10 then
			tremove(attackTimes)
			tremove(attackSpeeds)
		end

		mod:ExecuteSwap(isMainHand)			
		
		-- local amt, overkill, _, _, _, _, crit = ...
		-- if isMainHand then
		-- end
		-- print(("[%2.2f] %2.0f %s %s"):format(GetTime() - lastMHAttackTime, amt + overkill, crit and "Crit" or "", isMainHand and "Mainhand" or ""))
	end
end